#!/usr/bin/env python3
import datetime
import os
import pathlib
import re
import subprocess
import sys
import shlex
import shutil
from typing import List

from .utils import cprint, pad_string


class OpenerError(Exception):
    """Exception raised when command fails"""

    def __init__(self, error, message="ERROR: Failed to run"):
        self.error = error
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message} {self.error}"


def browse_base(user):
    """
    Browse a user's base directory with their desired program
    """
    base_dir = user.settings["base_dir"]
    opener = shlex.split(user.settings["file_manager"])
    if opener is None:
        raise OpenerError(user.files["conf_file"])
    opener.append(base_dir)
    try:
        open_process(opener)
    except OpenerError as err:
        cprint("red", err, out_file=sys.stderr)
        return 1


def check_mimetype(file: str):
    cmd = shlex.split(f"file --mime-type \"{file}\" -bL")
    result = return_cmd(cmd)
    return result["stdout"].strip("\n")


def check_zipped_dirs(dirs: List[pathlib.Path], zip_cmd: str, first_run: bool = True):
    recheck_flag = False
    for dir in dirs:
        if not dir.is_absolute():
            cprint("red", f"ERROR: Cannot fix zip: '{dir}.cbz'")
            continue
        parent = dir.parent.as_posix()
        cbz = pathlib.Path(dir.as_posix() + ".cbz")
        if not cbz.is_file():
            continue
        result = check_mimetype(cbz.as_posix())
        if result == "application/zip":
            continue
        cprint("yellow", f"Remaking broken zip '{cbz}'")
        if not recheck_flag:
            recheck_flag = True
        cbz.unlink()
        run_cmd([zip_cmd, parent, dir.as_posix()])
    if recheck_flag and first_run:
        check_zipped_dirs(dirs, zip_cmd, False)



def clean_dir_name(dir: pathlib.Path):
    pattern = r"([0-9]+\.?[0-9]*)"
    match = re.search(pattern, dir.name)
    if not match:
        return None
    new_name = pad_string(match.group(1))
    new = dir.parent.as_posix() + "/" + new_name
    return dir.rename(new)



def create_directory(base_dir, dir):
    """
    Create the given list of directories in base_dir
    """
    dir_path = pathlib.Path(base_dir, dir)
    if dir_path.is_dir():
        return True
    elif dir_path.is_file():
        return False
    dir_path.mkdir(parents=True)
    return True


def get_dirs(path: pathlib.Path):
    # list all directories under the given path
    directories = [
        entry
        for entry in path.glob("*")
        if entry.is_dir()
    ]
    return directories


def get_files(path: pathlib.Path):
    # list all files
    files = [
        file
        for file in path.glob("*")
        if file.is_file()
    ]
    return files


def get_new_count(dest, latest):
    path = pathlib.Path(dest)
    result = 0
    for file in get_files(path):
        try:
            name = file.name.split(".")[0]
            count = float(name)
        except (ValueError, IndexError):
            continue
        if count > latest:
            result += 1
    return result


def move_cbz(dest, path: pathlib.Path):
    target_dir = pathlib.Path(dest, path.name)
    for file in os.listdir(path):
        if not file.endswith("cbz"):
            continue
        dest_file = pathlib.Path(target_dir / file)
        if dest_file.exists() and dest_file.is_file():
            dest_file.unlink()
        elif dest_file.is_dir():
            raise OSError(f"Destination is a directory {dest_file}")
        shutil.move(file, target_dir)


def move_to_raw(base_dir, path: pathlib.Path):
    move_to_x(base_dir, "raw_manga", path)


def move_to_x(base_dir, x: str, path: pathlib.Path):
    target_dir = pathlib.Path(base_dir, x, path.name)
    dirs = get_dirs(path)
    for dir in dirs:
        target_path = target_dir / dir.name
        try:
            shutil.move(dir, target_path)
        except shutil.Error:
            cprint("red", f"ERROR: Cannot move '{dir}' to '{target_path}'")


def open_process(opener, out=subprocess.DEVNULL, err=subprocess.STDOUT):
    """Open a program with the given opener list"""
    try:
        subprocess.Popen(opener, stdout=out, stderr=err)
    except (subprocess.CalledProcessError, FileNotFoundError):
        raise OpenerError(opener)


def remove_directory(path: pathlib.Path):
    try:
        shutil.rmtree(path)
    except Exception as e:
        cprint("red", f"Error deleting directory {path}: {e}", out_file=sys.stderr)


def return_cmd(cmd):
    """
    Run a command and return the the output as a dict
    """
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    return {"stdout": stdout.decode("utf-8"), "stderr": stderr.decode("utf-8")}


def re_zip(user, args):
    if args.path is None:
        path = pathlib.Path().absolute()
    else:
        path = pathlib.Path(args.path)
    dirs = [dir for dir in get_dirs(path) if dir is not None]
    cprint("green", f"(Re)zipping '{path}'")
    zip_dirs(dirs, user.settings["zip_cmd"])
    if args.dir is None:
        base = user.settings["base_dir"]
    else:
        base = args.dir
    if args.move:
        move_cbz(base, path)


def run_cmd(cmd):
    subprocess.run(cmd)


def write_recent_updates(user, items):
    update_file = pathlib.Path(user.files["updates_file"])
    if update_file.is_file():
        update_file.replace(user.files["updates_bak_file"])
    update_file.touch()
    with open(update_file, "w") as file:
        file.write(f"Recent updates to library as of {datetime.datetime.now()}.\n")
        for item in items:
            title = item["title"]
            new_count = item["new_count"]
            file.write(
                f"{title} had {new_count} new chapter(s).\n"
            )



def zip_dirs(dirs: List[pathlib.Path], zip_cmd: str):
    # compress each directory using user cmd
    processes = []
    for dir in dirs:
        if not dir.is_absolute():
            cprint("red", f"ERROR: Cannot zip dir: '{dir}'")
            continue
        process = subprocess.Popen(
            [
                zip_cmd,
                dir.parent.as_posix(),
                dir.as_posix(),
                # "tar",
                # "--sort=name",
                # "-C",
                # dir_path,
                # "-caf",
                # f"{d}.zip",
                # chapter,
            ]
        )
        processes.append(process)
    for process in processes:
        process.communicate()



#!/usr/bin/env python3
import json
import re

# from .utils import join as utils_join
from os.path import isfile


def title_search(library, title):
    """
    First perform a simple search for the chapter by checking if it is
    listed in the user's library. This may not be 100% accurate if your library
    is not up-to-date, for that reason if the simple search does not return
    anything we do a deeper search to see if the file exists in your base_dir
    in a tracked show. If all of this fails we return None
    """
    for dir in library.keys():
        if dir == title:
            return dir
    return None


def main_search(user, chapter, track=False):
    """The main search loop"""
    with open(user.files["library_file"], "r") as data:
        library = json.load(data)
    # base_dir = user.settings["base_dir"]
    path = re.sub("(^.*)/.*$", r"\1", chapter)
    title = re.sub("^.*/", "", path)
    chapter = re.sub("^.*/", "", chapter)
    dir = title_search(library, title)
    if dir is None:
        return 1
    if track:
        track_chapter(dir, chapter, library, user)
    return 0


def track_chapter(dir, chapter, library, user):
    """
    Track an chapter in user's library by updating
    thier library and log file
    """
    log_file = user.files["log_file"]
    max = int(user.settings["max_history"])
    # Update library
    library[dir]["reading"] = chapter
    log = {}
    log[dir] = library[dir]
    if isfile(log_file):
        with open(log_file, "r") as data:
            old_log = json.load(data)
            if dir in old_log:
                old_log.pop(dir)
            if len(old_log) >= max:
                old_log.popitem()
            log.update(old_log)
    for dir in log.keys():
        log[dir] = library[dir]
    with open(log_file, "w+") as data:
        json.dump(log, data, indent=4)
    with open(user.files["library_file"], "w+") as data:
        json.dump(library, data, indent=4)
    return library



#!/usr/bin/env python3
import json
import os
import sys
from .utils import cprint, sort_updated_dirs, key_value_list, join
from .prompts import user_choice, InvalidCmdPrompt, InputError
from .system import open_process


def opener_cmd(item, user, path=None):
    """
    Return a command list to feed to system.open_process
    if a path to the item is given the two will be joined
    """

    opener_list = [user.settings["opener"]]
    if user.settings["extra_conf"]:
        opener_list.extend(["-c", user.config_dir])
    if path is not None:
        item = os.path.join(path, item)
    opener_list.append(item)
    return opener_list


def read_manga(user, latest=False, new=False):
    """
    Watch a show from the user's library
    """

    def ask_user(options, user, prompt):
        """
        Ask a user something and handle any errors, return False if the user
        reponds with nothing or error is caught
        """
        try:
            choice = user_choice(options=options, user=user, prompt=prompt)
        except (InvalidCmdPrompt, InputError, KeyboardInterrupt) as err:
            cprint("red", err, sys.stderr)
            return False
        if choice is None:
            return False
        return choice

    # If latest update log so we don't prompt for shows that don't exist anymore
    if latest:
        update_log(
            log_file=user.files["log_file"],
            library_file=user.files["library_file"],
        )
        with open(user.files["log_file"], "r") as data:
            library = json.load(data)
    else:
        with open(user.files["library_file"], "r") as data:
            library = json.load(data)

    # Get vars for prompting user and selecting the desired file to watch
    dirs, values = key_value_list(library)
    _, reading = key_value_list(values, search_key="reading")
    _, chapters = key_value_list(values, search_key="chapters")

    if new:
        ask_dirs = sort_updated_dirs(library)
    else:
        ask_dirs = dirs
    choice = ask_user(options=ask_dirs, user=user, prompt="Read: ")
    if not choice:
        return 1

    index = dirs.index(choice)
    path = join(user.settings["base_dir"], dirs[index])

    # If manga is does not have a recently read chapter ask user which chapter
    # to read
    if reading[index] is None:
        opts = chapters[index]
        chapter = ask_user(options=opts, user=user, prompt="Read")
        if not chapter:
            return 1
    else:
        chapter = reading[index]

    # Get the opener command and open manga
    cmd = opener_cmd(item=chapter, path=path, user=user)

    open_process(opener=cmd)

    return 0


def update_log(log_file, library_file):
    """
    Update user log before displaying
    """

    # Compare log file and library modification
    # times. Only update log if library has been
    # modified more recently than log
    log_time = os.path.getmtime(log_file)
    lib_time = os.path.getmtime(library_file)
    if lib_time < log_time:
        return None

    # Open log and library and load to objects
    with open(log_file, "r") as data:
        log = json.load(data)
    with open(library_file, "r") as data:
        library = json.load(data)

    # Create blank log
    lib_keys = list(library.keys())
    log_keys = list(log.keys())

    # Loop over log file and check if show is
    # still in library
    for dir in log_keys:
        if not dir in lib_keys:
            log.pop(dir)

    with open(log_file, "w+") as data:
        json.dump(log, data, indent=4)


# def zathura_cmd(item, user, path=None):
#     """
#     Return a command list to feed to system.open_process
#     if a path to the item is given the two will be joined
#     """

#     z_list = ["zathura"]
#     if user.settings["zathura_conf"]:
#         z_list.extend(["-c", user.config_dir])
#     if path is not None:
#         item = os.path.join(path, item)
#     z_list.append(item)
#     return z_list



#!/usr/bin/env python3
import json
import pathlib
import re
import sys
from os.path import join as os_join


def auto_title(dir_name, user_system):
    """
    Generate a title from dir_name
    e.g. /path/to/[Group] Show Season 01 (1080p) [Hash info]
    returns ' - path - to - Show Season 01 (1080p)'
    """

    title = re.sub(r"\s?\[[^]]*\]\s?", "", dir_name)
    if user_system == "Windows":
        title = re.sub("\\\\", " - ", title)
    else:
        title = re.sub(r"/", " - ", title)

    return title


def add_url(url, dir, user):
    """
    Associate the given url with given dir
    """
    lib = pathlib.Path(user.files["library_file"])
    dir_name = pathlib.Path(dir).name

    def backup_library():
        lib.replace(user.files["library_bak_file"])

    with open(user.files["library_file"], "r") as data:
        library = json.load(data)
    if dir_name in library.keys():
        library[dir_name]["url"] = url
        backup_library()
        with open(user.files["library_file"], "w+") as data:
            json.dump(library, data, indent=4)
            return 0
    return 1


def cprint(color: str, string, out_file=sys.stdout):
    """
    Red: \033[31m
    Green: \033[32m
    Yellow: \033[33m
    Blue: \033[34m
    Magenta: \033[35m
    Cyan: \033[36m
    """
    format = ""
    end = "\033[0m"
    if color.lower() == "red":
        format = "\033[31m"
    elif color.lower() == "green":
        format = "\033[32m"
    elif color.lower() == "yellow":
        format = "\033[33m"
    elif color.lower() == "blue":
        format = "\033[34m"
    elif color.lower() == "magenta":
        format = "\033[35m"
    elif color.lower() == "cyan":
        format = "\033[36m"
    else:
        print(string, file=out_file)
        return
    print(format + string + end, file=out_file)


def format_string(
    raw_string: str,
    args,
) -> str:
    format_dict = {
        "page": args.page,
        "track": args.track,
    }
    ret_str = raw_string
    format_keys = re.findall("{([^}]+)}", ret_str)
    for format_key in format_keys:
        # Check if item_key is a known key
        if format_key in format_dict.keys():
            replace_str = format_dict[format_key]
            ret_str = ret_str.replace(f"{{{format_key}}}", replace_str)
    return ret_str


def get_latest(entry):
    if entry["chapters"] != []:
        try:
            return float(entry["chapters"][-1].split(".")[0])
        except (ValueError, IndexError):
            return 0.0
    else:
        return 0.0


def is_updated(dest, data):
    path = pathlib.Path(dest)
    mtime = data["update_time"]
    return mtime < path.lstat().st_mtime


def join(a, b):
    return os_join(a, b)


def key_value_list(dic, search_key=None):
    """
    Take a dicionary and return two lists one for keys and one for values
    """
    # While it is easiest if dic is a true dict
    # it need not be. As long as the items in dic
    # _are_ true dicts then we can make do
    def psuedo_dic():
        for item in dic:
            if isinstance(item, dict):
                true_dic(item)

    def true_dic(d=dic):
        if search_key is None:
            keys.extend(d.keys())
            values.extend(d.values())
        else:
            for key, value in d.items():
                if key == search_key:
                    keys.append(key)
                    values.append(value)

    keys = []
    values = []
    if isinstance(dic, dict):
        true_dic()
    else:
        psuedo_dic()

    return keys, values


def merge_libraries(old_dict, new_dict):
    """Takes two dictionaries and merges them"""
    merged_dict = {}
    pop_key = None
    for new_key, new_val in new_dict.items():
        match = False
        for old_key, old_val in old_dict.items():
            if old_key == new_key:
                match = True
                old_val["chapters"] = new_val["chapters"]
                old_val["update_time"] = new_val["update_time"]
                merged_dict[old_key] = old_val
                pop_key = old_key
                break
        if match:
            old_dict.pop(pop_key, None)
            continue
        merged_dict[new_key] = new_val
    return merged_dict


def pad_string(string: str):
    # convert input string to float, then back to string with 1 decimal place
    converted_string = "{:.1f}".format(float(string))
    # pad the string with zeros on the left to make it 3 characters long
    padded_string = converted_string.zfill(5)
    return padded_string


def query_library(lib_file):
    ret = []
    with open(lib_file, "r") as data:
        library = json.load(data)
    for k, v in library.items():
        latest = get_latest(v)
        if v["url"] is None:
            continue
        ret.append({"url": v["url"], "title": k, "latest": latest})
    return ret


def sort_updated_dirs(library):
    update_dict = {}
    for k in library.keys():
        update_dict[library[k]["update_time"]] = k
    update_list = list(update_dict.keys())
    update_list.sort(reverse=True)
    return [update_dict[i] for i in update_list]



#!/usr/bin/env python3
import json
import multiprocessing
import pathlib
import re
import sys
import tempfile

from .build import update_library
from .media import opener_cmd
from .system import (
    check_zipped_dirs,
    clean_dir_name,
    create_directory,
    get_dirs,
    get_new_count,
    move_to_raw,
    remove_directory,
    return_cmd,
    run_cmd,
    write_recent_updates,
    zip_dirs,
)
from .utils import add_url, cprint, get_latest, is_updated, query_library


def make_item(user, args, max_str, url, title, latest):
    return {
        "user": user,
        "args": args,
        "max_str": max_str,
        "url": url,
        "title": title,
        "latest": latest,
        "updated": False,
        "new_count": 0
    }


def bulk_new(user, args):
    items = []
    if args.max_chapter is None:
        max_str = ""
    else:
        max_str = args.max_chapter
    if args.min_chapter is None:
        latest = 0.0
    else:
        try:
            latest = float(args.min_chapter)
        except ValueError:
            latest = 0.0
    for url in args.add:
        title = get_title(url)
        if title is None:
            cprint(
                "red",
                f"Could not get title, skipping url: {url}",
                out_file=sys.stderr,
            )
            continue
        if not create_directory(
            base_dir=user.settings["base_dir"],
            dir=title,
        ):
            cprint(
                "red",
                f"Could not create directory '{title}' for url: {url}",
                out_file=sys.stderr,
            )
            continue
        items.append(make_item(user, args, max_str, url, title, latest))
    pool = multiprocessing.Pool()
    results = pool.map_async(download_latest, items)
    results.get()
    cprint("green", "Updating library")
    update_library(user, args)
    for item in items:
        update_urls(item)
    return 0


def bulk_latest(user, args):
    data = query_library(user.files["library_file"])
    items = []
    if args.max_chapter is None:
        max_str = ""
    else:
        max_str = args.max_chapter
    for item in data:
        items.append(
            make_item(
                user, args, max_str, item["url"], item["title"], item["latest"]
            )
        )
    pool = multiprocessing.Pool()
    results = pool.map_async(download_latest, items)
    results.get()
    updated = []
    with open(user.files["library_file"], "r") as data:
        library = json.load(data)
    cprint("green", "Updating library")
    update_library(user, args)
    for title, data in library.items():
        dest = user.settings["base_dir"] + f"/{title}"
        if is_updated(dest, data):
            new_count = get_new_count(dest, get_latest(data))
            updated.append(
                {
                    "title": title,
                    "new_count": new_count
                }
            )
            cprint("cyan", f"Was updated: {title}")
        else:
            cprint("magenta", f"Not updated: {title}")
    if len(updated) != 0:
        cprint("green", f"Documenting updates in '{user.files['updates_file']}'")
        write_recent_updates(user, updated)
    return 0


def download_latest(item):
    user = item["user"]
    # # args = item[1]
    max_str = item["max_str"]
    url = item["url"]
    title = item["title"]
    latest = item["latest"]
    base_dir = user.settings["base_dir"]
    dest = f"{base_dir}/{title}"
    gdl_config_path = user.files["gdl_file"]
    if max_str == "":
        max_count = 9999
    else:
        try:
            max_count = latest + float(max_str)
        except ValueError:
            max_count = latest + 1
    cmd = [
        "gallery-dl",
        "-f", "/O",
        "--config", gdl_config_path,
        "-d", f"{dest}/gallery-dl",
        "--chapter-filter", f"{latest} < chapter <= {max_count}",
        "-o", "cookies.PHPSESSID=aaaaaaaaaaaaaaaaaaaaaaaaaa",
        "--exec-after", f'mv {{_directory}} "{dest}"/',
        url
    ]
    cprint("green", f"Downloading {title}: {url}")
    run_cmd(cmd)
    storage_path = pathlib.Path(dest + "/gallery-dl")
    if storage_path.is_dir():
        remove_directory(storage_path)
        process_dir(base_dir, dest, user)
    cprint("green", f"Finished {title}: {url}")


def get_title(url):
    result = return_cmd(["gallery-dl", "-K", url])
    if result["stderr"] != "":
        return None
    pattern = r"manga\n\s+(.+)\n"
    match = re.search(pattern, result["stdout"])
    if match:
        return match.group(1)
    else:
        return None


def process_dir(base_dir, dest, user):
    path = pathlib.Path(dest)
    dirs = []
    cprint("yellow", f"Cleaning: {path}")
    for dir in get_dirs(path):
        result = clean_dir_name(dir)
        if result is not None:
            dirs.append(result)
    cprint("yellow", f"Zipping: {path}")
    zip_dirs(dirs, user.settings["zip_cmd"])
    cprint("yellow", f"Checking zips: {path}")
    check_zipped_dirs(dirs, user.settings["zip_cmd"])
    cprint("yellow", f"Moving raws: {path}")
    move_to_raw(base_dir, path)


def update_urls(item):
    user = item["user"]
    url = item["url"]
    title = item["title"]
    base_dir = user.settings["base_dir"]
    dest = f"{base_dir}/{title}"
    cprint("yellow", f"Associating {url} with {title}")
    add_url(url, dest, user)


def view_chapter(user, args):
    with tempfile.TemporaryDirectory() as tmpdir:
        path = pathlib.Path(tmpdir)
        url = args.view
        if args.chapter is not None:
            try:
                chapter = float(args.chapter)
            except ValueError:
                chapter = 1.0
        else:
            chapter = 1.0
        gdl_config_path = user.files["gdl_file"]
        cmd = [
            "gallery-dl",
            "-f", "/O",
            "--config", gdl_config_path,
            "-d", f"{path}/gallery-dl",
            "--chapter-filter", f"{chapter} <= chapter <= {chapter}",
            "-o", "cookies.PHPSESSID=aaaaaaaaaaaaaaaaaaaaaaaaaa",
            "--exec-after", f'mv {{_directory}} "{path}"/',
            url
        ]
        run_cmd(cmd)
        storage_path = pathlib.Path(path / "gallery-dl")
        if not storage_path.is_dir():
            cprint("red", "ERROR: Download failed.")
            return 1
        remove_directory(storage_path)
        dirs = []
        for dir in get_dirs(path):
            result = clean_dir_name(dir)
            if result is not None:
                dirs.append(result)
        if len(dirs) == 0:
            cprint("red", "ERROR: Did not download anything.")
            return 1
        elif len(dirs) != 1:
            cprint("red", "ERROR: Downloaded too many chapters.")
            return 1
        zip_dirs(dirs, user.settings["zip_cmd"])
        check_zipped_dirs(dirs, user.settings["zip_cmd"])
        dir = dirs[0]
        file = f"{dir}.cbz"
        view_cmd = [user.settings["opener"]]
        if user.settings["extra_conf"]:
            view_cmd.extend(["-c", user.settings["extra_conf"]])
        view_cmd.append(file)
        run_cmd(view_cmd)



#!/usr/bin/env python3
from promptx import PromptX
from .utils import cprint


class InvalidCmdPrompt(Exception):
    """Exception raised when user has invalid command prompt"""

    def __init__(self, error, message="ERROR: prompt_cmd not recognized"):
        self.error = error
        self.message = message

    def __str__(self):
        return f'{self.message} "{self.error}"'


class InputError(Exception):
    """Exception fzf or dmenu prompt fails"""

    def __init__(self, error, message="ERROR: Could not get user input", prefix=""):
        self.error = error
        self.prefix = prefix
        self.message = message

    def __str__(self):
        return f"{self.prefix}{self.message}{self.error}"


def fzf_auto(options, prompt):
    choice = None
    try:
        cprint("green", f"Auto title: {options[0]}")
        choice = input(prompt)
        if choice == "":
            choice = options[0]

    except KeyboardInterrupt:
        raise InputError("User interruption...stopping", message="", prefix="\n")

    return choice


def user_choice(
    options,
    user,
    prompt,
    auto_title=False,
):
    """
    Give user a prompt to choose from a list of options.  Uses dmenu, fzf, or
    rofi
    """
    cmd = user.settings["prompt_cmd"]
    cmd_args = user.settings["prompt_args"]
    p = PromptX(prompt_cmd=cmd)
    if cmd == "dmenu" and cmd_args == "":
        cmd_args = "-l 20 -i"
    elif cmd == "fzf" and auto_title:
        choice = fzf_auto(options, prompt)
        return choice

    choice = p.ask(
        options=options,
        prompt=prompt,
        additional_args=cmd_args,
    )

    try:
        return choice[0]
    except IndexError:
        return []



#!/usr/bin/env python3
import shlex
import sys


from .build import build_main, update_library
from .config import NoBaseDir, NoBaseDirExists, get_user_settings
from .g_dl import bulk_new, bulk_latest, view_chapter
from .media import read_manga
from .options import get_opts
from .search import main_search
from .system import browse_base, open_process, re_zip
from .utils import add_url, cprint, format_string


__license__ = "GPL-v3.0"
__program__ = "mlm"


def process_opts(user, args):
    """
    Opts handler for main
    """
    if args.build:
        cprint("green", "Building")
        return build_main(user, args)
    elif args.browse:
        return browse_base(user)
    elif args.open is not None:
        cprint("green", "Opening the given file")
    elif args.search is not None:
        return main_search(user, chapter=args.search)
    elif args.track is not None:
        main_search(user, chapter=args.track, track=True)
        cmd = shlex.split(format_string(user.settings["track_cmd"], args))
        # cmd = shlex.split(f"znp -e 'quit' -g '{args.page}' '{args.track}'")
        return open_process(opener=cmd)
    elif args.update:
        cprint("green", "Updating library")
        return update_library(user, args)
    elif args.read or args.latest:
        return read_manga(user, latest=args.latest, new=args.new)
    elif args.url is not None and args.dir is not None:
        return add_url(url=args.url, dir=args.dir, user=user)
    elif args.add is not None:
        return bulk_new(user=user, args=args)
    elif args.download_new:
        return bulk_latest(user=user, args=args)
    elif args.view is not None:
        return view_chapter(user, args)
    elif args.zip:
        return re_zip(user=user, args=args)
    else:
        cprint("green", "Updating library")
        return update_library(user, args)


def main():
    """
    Command line application to view and track media
    """
    # Set and get command line args
    args = get_opts(__program__)

    try:
        # Creates a UserSettings object. This will be used by various function
        # to access file paths, settings, filters, and command line args
        user, args = get_user_settings(program=__program__, args=args)
    except (NoBaseDir, NoBaseDirExists) as err:
        cprint("red", err.message, out_file=sys.stderr)
        return 1

    # Execute the appropriate function based on command line options
    return process_opts(user, args=args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
import argparse


def get_opts(prog_name="mlm"):
    parser = argparse.ArgumentParser(
        prog=prog_name,
        description="""Track and watch your media library""",
        allow_abbrev=False,
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-a",
        "--add",
        metavar="URL",
        action="append",
        help="""
        Download manga and track URL. Can give the --max-chapter flag to limit the
        number of chapters downloaded.
        """,
    )
    group.add_argument(
        "-b",
        "--build",
        action="store_true",
        help="""Build the LIBRARY file. Use -i to build this file interactively.""",
    )
    group.add_argument(
        "-B",
        "--browse",
        action="store_true",
        help="""
        Browse your BASE_DIR. Set your file manager of choice in the mlm.conf
        like so:
            FILE_MANAGER="st -e lf"
        TUI file managers like lf, nnn, ranger,
        etc. typically must be started in a terminal emulator.
        """,
    )
    group.add_argument(
        "-D",
        "--download-new",
        action="store_true",
        help="""
        Download new chapters of manga in your library.
        A title must have an associated url.
        By default will download all chapters after your most recent chapter.
        Use -m, --max-chapters N to limit new downloads to N more than your most
        recent chapter.
        """,
    )
    group.add_argument(
        "-l",
        "--latest",
        action="store_true",
        help="""List your latest tracked manga to resume reading.""",
    )
    group.add_argument(
        "-o",
        "--open",
        metavar="FILE",
        help="""
        This is a sort of wrapper for zathura which will also track the anime
        when closed.
        """,
    )
    group.add_argument(
        "-r",
        "--read",
        action="store_true",
        help="""
        List all titles in your library. If the chosen title is being
        tracked then the last tracked chapter will be shown in zathura.
        If this is a title that has no tracked chapter then a list
        of all chapters found in the title's dir
        will be listed for you to choose from.
        """,
    )
    group.add_argument(
        "-s",
        "--search",
        metavar="FILE",
        help="""
        Take a file path and check if it is in your BASE_DIR. Returns 0 if
        successful or 1 if the file is not in in your BASE_DIR.
        """,
    )
    group.add_argument(
        "-t",
        "--track",
        metavar="FILE",
        help="""
        This will track the given FILE. It is best if the the given FILE is the
        full path to the file. If FILE is a relative path then mlm will attempt
        to find this file in one of the dirs in your library.  If this is not
        possible you will be informed. The most likely reason that this would
        happen is that the file is in a dir that has been filtered or it is a
        new dir that has not been added to your library. In this case update
        your library and attempt to track again. FILE should be a file somewhere
        in your BASE_DIR otherwise it will be impossible to track.
        """,
    )
    group.add_argument(
        "-u",
        "--update",
        action="store_true",
        help="""
        This will update your library file to match any new or renamed
        directories. If the directory name has not changed your tracked episodes
        will carry over. If this is not a new directory but one that has been
        renamed it will be treated as new and no episode tracking will be
        carried over.
        """,
    )
    group.add_argument(
        "-U",
        "--url",
        metavar="URL",
        action="store",
        help="""
        URL to associate with DIR.
        """,
    )
    group.add_argument(
        "-v",
        "--view",
        metavar="URL",
        action="store",
        help="""
        This will download the first chapter of URL and open it up in your viewer.
        By default this is chapter 1 or the chapter number indicated by '--chapter N'.
        The downloaded chapter will be deleted after the viewer command exits,
        regardless of exit status.
        """,
    )
    group.add_argument(
        "-z",
        "--zip",
        action="store_true",
        help="""
        (Re)zip the PATH given by --path or the $PWD if not supplied. This will
        create a cbz file for each dir matching the following regex:
        ^.*[0-9]\\.[0-9]$
        """,
    )
    parser.add_argument(
        "-c",
        "--chapter",
        metavar="N",
        action="store",
        help="""
        Use with --view if you want to some other chapter besides the first one.
        """,
    )
    parser.add_argument(
        "-C",
        "--clean",
        action="store_true",
        help="""
        Use this flag with -u if you do not want to keep a backup of your library.
        """,
    )
    parser.add_argument(
        "-d",
        "--dir",
        metavar="DIR",
        action="store",
        help="""
        DIR to use for other options like --url, and --move
        """,
    )
    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="""
        Use when building or updating your library.  This will make the process
        interactive. By default mlm tries to set the title of a dir to something
        sane.  See the man page or github for more detailed info about this. If
        you prefer to set the title value yourself then use this flag. If run in
        a terminal you will see what the auto generated title would be if you do
        not enter anything into the dmenu prompt by pessing ESC
        """,
    )
    parser.add_argument(
        "-m",
        "--max-chapter",
        metavar="N",
        action="store",
        help="""
        Download a max of N chapters. If downloading with --add this will
        be an absolute max. If downloading with --download-new this will be
        N more than the most recent chapter in your library.
        """,
    )
    parser.add_argument(
        "-M",
        "--min-chapter",
        metavar="N",
        action="store",
        help="""
        Download chapters starting at N. This is only used when using --add
        """,
    )
    parser.add_argument(
        "-n",
        "--new",
        action="store_true",
        help="""
        Use with --read to list manga in order of update order.
        """,
    )
    parser.add_argument(
        "--move",
        action="store_true",
        help="""
        Used with --zip. If DIR is given the zipped files will be placed in
        """,
    )
    parser.add_argument(
        "-p",
        "--page",
        action="store",
        type=int,
        default=-1,
        help="""
        Used for better accuracy when tracking a file.
        """,
    )
    parser.add_argument(
        "-P",
        "--path",
        metavar="PATH",
        action="store",
        help="""
        The PATH to --zip.
        """,
    )
    args = parser.parse_args()
    return args



#!/usr/bin/env python3
import json
import mimetypes
import os
import pathlib
from promptx import (
    PromptXCmdError,
    PromptXError,
    PromptXSelectError,
)
import re
import sys
from .utils import cprint, merge_libraries
from .prompts import user_choice, InputError, InvalidCmdPrompt


def filter_directory(dir, filters):
    """
    Filter a directory based on a filter list.
    Optionally recursive
    """
    escape_dir = re.escape(dir)
    for item in filters:
        escape_item = re.escape(item)
        if re.search(f"{escape_item}$", escape_dir):
            return True
    return False


def is_cbz(file: str):
    """
    Test if the given file is a video
    """
    return file.endswith("cbz")


def pre_build_check(user, prompt=True):
    """Check for existing user library"""

    # Create a config Path object
    library = pathlib.Path(user.files["library_file"])

    def backup_library():
        library.replace(user.files["library_bak_file"])

    # prompt_user only runs when prompt is True
    def prompt_user():
        # Ask user if they want to backup their library
        # This prompt should only run when building
        # User should not be prompted when updating
        choice = user_choice(["Yes", "No"], user, prompt="Backup library? ")
        if choice == "Yes":
            cprint("green", "Backing up library")
            backup_library()
        else:
            cprint("yellow", "Overwriting library")

    # Check if the file exists, if not create it
    if not library.is_file():
        library.touch()
        # When building this should return True
        # When updating this should return False
        return prompt
    elif prompt:
        if library.stat().st_size != 0:
            prompt_user()
        return True
    else:
        if library.stat().st_size != 0:
            backup_library()
            return True
        else:
            return False


def update_library(user, args):
    """Update a user's library"""

    # If the user's library needs to be created or is of size
    # '0' then just build the library, else update
    if not pre_build_check(user, prompt=False):
        return build_main(user, args)
    # Get a fresh library
    new_library = build_main(user, args, write=False)
    # Get the old library
    with open(user.files["library_bak_file"], "r") as old_data:
        old_library = json.load(old_data)
    # Merge the two libraries
    merged_library = merge_libraries(old_library, new_library)
    # Write updated library
    with open(user.files["library_file"], "w") as new_data:
        json.dump(
            merged_library,
            new_data,
            indent=4,
        )
    return 0


def walk_path(user, args):
    """
    Walk the base dir and return a dict with valid dirs and episodes
    Create a valid_dirs dict. This will store the raw dirs
    as the key for a show. Here is the structure
    {
        "path/to/manga": {
            "reading": "currently reading",
            "url": "https://mangadex.org/title/4edaaece-5b66-4fd5-a7e2-01f7a017ccdb/gekkan-shoujo-nozaki-san"
        }
    }
    """
    chapters = {}
    dirs = []
    valid_dirs = {}
    skip_dir = True
    base_dir = user.settings["base_dir"]
    contents = os.scandir(base_dir)
    for item in contents:
        if (
            filter_directory(item.name, user.stored["filters_file"])
            or not item.is_dir()
        ):
            continue
        chapters[item.name] = []
        files = os.scandir(item.path)
        for file in files:
            if not file.is_file():
                continue
            if is_cbz(file=file.path):
                if skip_dir:
                    skip_dir = False
                chapters[item.name].append(file.name)
        if skip_dir:
            continue
        chapters[item.name].sort()
        dirs.append(item.name)
    dirs.sort()
    for dir in dirs:
        valid_dirs[dir] = {}
        valid_dirs[dir]["reading"] = None
        valid_dirs[dir]["url"] = None
        valid_dirs[dir]["chapters"] = chapters[dir]
        mtime = pathlib.Path(base_dir, dir).lstat().st_mtime
        valid_dirs[dir]["update_time"] = mtime
    return valid_dirs


def build_main(user, args, write=True):
    """
    The main build loop for a user's library
    """

    try:
        pre_build_check(user)
    except (
        InvalidCmdPrompt,
        InputError,
        KeyboardInterrupt,
        PromptXCmdError,
        PromptXError,
        PromptXSelectError,
    ) as err:
        cprint("red", err, out_file=sys.stderr)
        return 1
    try:
        valid_dirs = walk_path(user, args)
    except (
        InvalidCmdPrompt,
        InputError,
        KeyboardInterrupt,
        PromptXCmdError,
        PromptXError,
        PromptXSelectError,
    ) as err:
        cprint("red", err, out_file=sys.stderr)
        return 1
    if write:
        with open(user.files["library_file"], "w") as data:
            json.dump(valid_dirs, data, indent=4)
    else:
        return valid_dirs
    return 0



#!/usr/bin/env python3
import os
from loadconf import Config


class NoBaseDirExists(Exception):
    """Exception raised when user has a base directory set
    but it does not exist.
    """

    def __init__(self, conf_file, message="ERROR: Base directory is set to"):
        self.file = conf_file
        self.message = message

    def __str__(self):
        return f'{self.message} "{self.file}" which does not exist'


class NoBaseDir(Exception):
    """Exception raised when user has not set a base directory"""

    def __init__(self, conf_file, message="ERROR: Base directory not set in"):
        self.file = conf_file
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} "{self.file}"'


def get_user_settings(program, args):
    # Create user object to read files and get settings
    user = Config(program=program)
    # Define some basic settings, files, etc.
    user_settings = {
        "base_dir": None,
        "debug": False,
        "extra_conf": False,
        "file_manager": None,
        "language": "",
        "max_history": 999,
        "not_found": False,
        "opener": "sioyek",
        "prompt_cmd": "fzf",
        "prompt_args": "",
        "track_cmd": "znp -e 'quit' -g {page} {track}",
        "zip_cmd": "mlm-zip",
    }
    config_files = {
        "conf_file": "mlm.conf",
        "log_file": "log.json",
        "filters_file": "filters.conf",
        "library_file": "library.json",
        "library_bak_file": "library.json.bak",
        "gdl_file": "gdl_config.json",
        "updates_file": "updates.txt",
        "updates_bak_file": "updates.txt.bak",
    }
    files = [
        "conf_file",
        "filters_file",
        "library_file",
    ]
    settings = list(user_settings.keys())
    # Fill out user object
    user.define_settings(settings=user_settings)
    user.define_files(user_files=config_files)
    user.create_files(create_files=files)
    user.read_conf(user_settings=settings, read_files=["conf_file"])
    user.store_files(files=["filters_file"])
    # Check that the required settings are defined
    if user.settings["base_dir"] is None:
        raise NoBaseDir(conf_file=user.settings["base_dir"])
    elif not os.path.isdir(user.settings["base_dir"]):
        raise NoBaseDirExists(conf_file=user.settings["base_dir"])

    return user, args
